<?php
  error_reporting(E_ALL);
  require_once(dirname(__FILE__) . DIRECTORY_SEPARATOR . 'sha1.php');
  require_once(SM_PATH . 'class/deliver/Deliver.class.php');

  define('DOMAINKEYS_HEADER_TAG', 'DomainKey-Signature');
  define('DOMAINKEYS_SENDER_HEADER', 'sender');
  define('DOMAINKEYS_FROM_HEADER', 'from');
  define('DOMAINKEYS_DEFAULT_SELECTOR', 'beta');
  define('DOMAINKEYS_QUERY_METHOD', 'dns');
  define('DOMAINKEYS_ALGORITHM', 'rsa-sha1');
  define('DOMAINKEYS_CANONICALIZATION_ALGORITHM', 'nofws');
  define('DOMAINKEYS_HEADER_FIELDS_LIST',
         'received,message-id,date,from,reply-to,to,subject,mime-version,content-type,content-transfer-encoding');

  define('DOMAINKEYS_DNS_NAMESPACE', '_domainkey');
  define('DOMAINKEY_STATUS_GOOD', 0);
  define('DOMAINKEY_STATUS_BAD', 1);
  define('DOMAINKEY_STATUS_NOKEY', 2);
  define('DOMAINKEY_STATUS_REVOKED', 3);
  define('DOMAINKEY_STATUS_NOSIGNATURE', 4);
  define('DOMAINKEY_STATUS_BADFORMAT', 5);
  define('DOMAINKEY_STATUS_NONPARTICIPANT', 6);

  define('RSASHA1_PREFIX', pack('H*', '3021300906052b0e03021a05000414'));

  function last($str, $count)
  {
    return substr($str, strlen($str) - $count);
  }

  class Deliver_DomainKeys extends Deliver
  {
    var $canonicalization_algorithm;
    var $domain;
    var $body = false;
    var $allowed = array();
    var $used = array();
    var $line = '';
    var $last = '';
    var $sha1;
    var $header = array();

    function initStream($domain, $allowed_headers=DOMAINKEYS_HEADER_FIELDS_LIST, $canonicalization_algorithm='nofws')
    {
      $this->canonicalization_algorithm = strtolower(trim($canonicalization_algorithm));
      $this->domain = $domain;
      $this->body = false;
      $this->allowed = $this->used = array();
      if($allowed_headers != '') foreach(explode(',', $allowed_headers) as $h)
      {
        $h = strtolower(trim($h));
        $this->allowed[$h] = $h;
      }
      $this->line = $this->last = '';
      $this->sha1 = sha1_create();
      return true;
    }

    function parse_header($header)
    {
      if(is_null($header) || trim($header) == '') return false;
      $ret = array();
      foreach(explode(';', $header) as $entry)
      {
        $entry = trim($entry);
        $value = trim(substr($entry, 2));
        switch(substr($entry, 0, 2))
        {
          case 'a=': $ret['algorithm'] = $value; break;
          case 'b=': $ret['signature_data'] = $value; break;
          case 'c=': $ret['canonicalization_algorithm'] = $value; break;
          case 'd=': $ret['domain'] = $value; break;
          case 'h=': $ret['header_list'] = strtr($value, ':', ','); break;
          case 'q=': $ret['query_method'] = $value; break;
          case 's=': $ret['selector'] = $value; break;
//        default: syslog(LOG_WARNING, "Unknown tag: " . $entry); break;
        }
      }
      return $ret;
    }

    function preWriteToStream($content)
    {
      for($i = 0; $i < strlen($content); $i++)
      {
        if($this->canonicalization_algorithm == 'nofws' && $content{$i} == "\r") {}
        elseif($this->body)
        {
          if($content{$i} == "\n")
            $this->line .= "\r\n";
          else
          {
            if(trim($this->line, "\r\n") != '')
            {
              if($this->canonicalization_algorithm == 'nofws')
              {
                $this->line = str_replace(' ', '', $this->line);
                $this->line = str_replace("\t", '', $this->line);
              }
              sha1_update($this->sha1, $this->line);
              $this->last = last($this->line, 1);
              $this->line = '';
            }
            if($this->canonicalization_algorithm == 'nofws')
            {
              if($content{$i} != ' ' && $content{$i} != "\t" && $content{$i} != "\r")
              {
                if($this->line != '')
                {
                  sha1_update($this->sha1, $this->line);
                  $this->last = last($this->line, 1);
                  $this->line = '';
                }
                if($content{$i} == "\n")
                {
                  sha1_update($this->sha1, "\r");
                }
                sha1_update($this->sha1, $content{$i});
                $this->last = $content{$i};
              }
            }
            else // canonicalization_algorithm = simple
            {
              if($this->line != '')
              {
                $this->line = str_replace("\r\n", "\n", $this->line);
                $this->line = str_replace("\n", "\r\n", $this->line);
                sha1_update($this->sha1, $this->line);
                $this->line = '';
              }
              sha1_update($this->sha1, $content{$i});
              $this->last = $content{$i};
            }
          }
        }
        elseif($this->line == "\n")
        {
          $this->body = true;
          $this->line = "\r\n" . $content{$i};
        }
        elseif(last($this->line, 1) == "\n")
        {
          if($content{$i} == ' ' || $content{$i} == "\t")
          {
            $this->line .= $content{$i};
          }
          else
          {
            list($key, $value) = explode(':', $this->line, 2);
            $key = strtolower(trim($key));
            if($this->canonicalization_algorithm == 'nofws')
              $this->line = str_replace(' ', '', strtr($this->line, "\t\n", '  ')) . "\r\n";
            else
              $this->line = str_replace("\n", "\r\n", $this->line);

            if(isset($this->allowed[$key]))
            {
              $this->used[$key] = $key;
              sha1_update($this->sha1, $this->line);
              $this->last = last($this->line, 1);
            }

            if($key == strtolower(DOMAINKEYS_HEADER_TAG))
            {
              $header = $this->parse_header($this->line);
              $this->initStream($header['domain'], $header['header_list'], $header['canonicalization_algorithm']);
              $this->header = $header;
              $this->header[DOMAINKEYS_HEADER_TAG] = $this->line;
            }
            elseif($key == DOMAINKEYS_SENDER_HEADER || $key == DOMAINKEYS_FROM_HEADER)
            {
              $this->header[$key] = $value;
            }
            $this->line = $content{$i};
          }
        }
        else
        {
          $this->line .= $content{$i};
        }
      }
    }

    function finalizeStream()
    {
      if($this->canonicalization_algorithm == 'nofws')
      {
        $this->line = str_replace(' ', '', strtr($this->line, "\t\r\n", '   '));
        if($this->line != '')
        {
          sha1_update($this->sha1, $this->line);
          $this->last = last($this->line, 1);
        }
      }
      else // canonicalization_algorithm = simple
      {
        $this->line = str_replace("\r\n", "\n", $this->line);
        $this->line = str_replace("\n", "\r\n", rtrim($this->line, "\n"));
        if($this->line != '')
        {
          sha1_update($this->sha1, $this->line);
          $this->last = last($this->line, 1);
        }
      }

      if($this->last == '')
      {
        if(rtrim($this->line, "\r\n") == '')
        {
          sha1_update($this->sha1, "\r\n");
        }
      }
      elseif($this->last != "\n")
      {
        sha1_update($this->sha1, "\r\n");
      }
    }

    function sender()
    {
      $sender = null;
      if(isset($this->header[DOMAINKEYS_FROM_HEADER]))
      {
        $sender = (isset($this->header[DOMAINKEYS_SENDER_HEADER])) ? $this->header[DOMAINKEYS_SENDER_HEADER] : $this->header[DOMAINKEYS_FROM_HEADER];
        $p = explode('@', trim($sender));
        $p = explode(' ', strtr($p[count($p)-1], "\t\r\n<>", '     '));
        $sender = strtolower($p[0]);
      }
      return $sender;
    }

    function is_subdomain_of($a, $b)
    {
      $a = strtolower(strrev($a)) . '.';
      $b = strtolower(strrev($b)) . '.';
      return (substr($b, 0, strlen($a)) == $a) ? true : false;
    }

    function sign($selector=DOMAINKEYS_DEFAULT_SELECTOR, $key_file=null)
    {
      if(is_null($key_file)) $key_file = "/etc/domain_keys/{$this->domain}.private.key";
      if(!file_exists($key_file)) return "X-Domain-Keys-Error: Missing key file $key_file";
      $private_key = openssl_get_privatekey(file_get_contents($key_file));

      if(isset($this->header[DOMAINKEYS_HEADER_TAG])) return "X-Domain-Keys-Error: Already signed";

      $sender = $this->sender();
      if(!$this->is_subdomain_of($sender, $this->domain)) return "X-Domain-Keys-Error: Not Authorized ($sender, {$this->domain})";

      $signature = '';
      $sha1 = sha1_finish($this->sha1);

      openssl_private_encrypt(RSASHA1_PREFIX . $sha1, $signature, $private_key);

      $ret = DOMAINKEYS_HEADER_TAG . ': ';
      $ret .= 'a=' . DOMAINKEYS_ALGORITHM . '; ';
      $ret .= 'c=' . DOMAINKEYS_CANONICALIZATION_ALGORITHM . '; ';
      $ret .= "s=$selector; d={$this->domain}; ";
      $ret .= 'q=' . DOMAINKEYS_QUERY_METHOD . '; ';
      $ret .= 'h=' . join(':', $this->used) . '; ';
      $ret .= 'b=' . base64_encode($signature) . ';';
      return $ret;
    }

    function get_dns($selector)
    {
      $txt = dns_get_record($selector . '.' . DOMAINKEYS_DNS_NAMESPACE . '.' . $this->domain, DNS_TXT);
      if(!isset($txt[0]['txt']) || trim($txt[0]['txt']) == '') return false;

      $ret = array();

      foreach(explode(';', $txt[0]['txt']) as $entry)
      {
        $entry = trim($entry);
        $value = trim(substr($entry, 2));
        switch(substr($entry, 0, 2))
        {
          case 'p=': $ret['public_key_data'] = $value; break;
          case 't=': $ret['testing_mode'] = ($value == 'y') ? true : false; break;
          case 'r=': $ret['reporting_email_address'] = $value; break;
          case 'o=': $ret['outbound_signing_policy'] = $value; break;
          case 'n=': $ret['notes'] = $value; break;
          case 'k=': $ret['key_type'] = $value; break;
          case 'g=': $ret['granularity_of_the_key'] = intval($value);
//          default:  syslog(LOG_WARNING, "Unknown tag: " . $entry); break;
        }
      }
      return $ret;
    }

    function check()
    {
      if(!isset($this->header[DOMAINKEYS_HEADER_TAG])) return DOMAINKEY_STATUS_NOSIGNATURE;
      if(!$this->is_subdomain_of($this->domain, $this->sender())) return DOMAINKEY_STATUS_BADFORMAT;

      $dns = $this->get_dns();
      if($dns === false || !isset($dns['public_key_data'])) return DOMAINKEY_STATUS_NOKEY;

      $sha1 = sha1_finish($this->sha1);

      $cert = "-----BEGIN PUBLIC KEY-----\n";
      for($i = 0; $i < strlen($dns['public_key_data']); $i += 64)
        $cert .= substr($dns['public_key_data'], $i, 64) . "\n";
      $cert .= "-----END PUBLIC KEY-----\n";
      $key = openssl_get_publickey($cert);

      $hash = '';
      openssl_public_decrypt(base64_decode($this->header['signature_data']), $hash, $key);

      $hash = substr($hash, strlen(RSASHA1_PREFIX));
      if($hash == $sha1) return DOMAINKEY_STATUS_GOOD;

      if($dns['testing_mode'])
      {
//        mail($dns['reporting_email_address'], 'ERROR', $mail);
        return DOMAINKEY_STATUS_GOOD;
      }

      if($dns['outbound_signing_policy'] == '~') return DOMAINKEY_STATUS_GOOD;
      if($dns['outbound_signing_policy'] == '-') return DOMAINKEY_STATUS_BAD;

      return DOMAINKEY_STATUS_BAD;
    }
  }
?>
