<?php
// SHA1
// David Eder <david@eder.us>
// 
// based on code by Steve Reid <steve@edmweb.com>
//

function sha1_safe_add($x, $y)
{
  $lsw = ($x & 0xFFFF) + ($y & 0xFFFF);
  $msw = ($x >> 16) + ($y >> 16) + ($lsw >> 16);
  return ($msw << 16) | ($lsw & 0xFFFF);
}

function sha1_zero_fill($a, $b)
{
  $bin = decbin($a);
  $len = strlen($bin);
  $bin = ($len < $b) ? 0 : substr($bin, 0, strlen($bin) - $b);
  return bindec(str_pad($bin, $len, '0', STR_PAD_LEFT));
}

function sha1_rol($value, $bits)
{
  return ($value << $bits) | sha1_zero_fill($value, 32 - $bits);
}

function sha1_blk0(&$block, $i)
{
  return ($block[$i] = (sha1_rol($block[$i],24) & 0xFF00FF00) | (sha1_rol($block[$i], 8) & 0x00FF00FF));
}

function sha1_blk(&$block, $i)
{
  $block[$i & 15] = $block[($i+13) & 15] ^ $block[($i+8) & 15] ^ $block[($i+2)&15] ^ $block[$i&15];
  return ($block[$i & 15] = sha1_rol($block[$i & 15], 1));
}

function sha1_R0(&$block, &$data, $v, $w, $x , $y, $z, $i)
{
  $a = sha1_safe_add(($data[$w] & ($data[$x] ^ $data[$y])) ^ $data[$y], sha1_blk0($block, $i));
  $a = sha1_safe_add($a, 0x5A827999);
  $a = sha1_safe_add($a, sha1_rol($data[$v] ,5));
  $data[$z] = sha1_safe_add($data[$z], $a);
  $data[$w] = sha1_rol($data[$w], 30);
}

function sha1_R1(&$block, &$data, $v, $w, $x, $y, $z, $i)
{
  $a = sha1_safe_add(($data[$w] & ($data[$x] ^ $data[$y])) ^ $data[$y], sha1_blk($block, $i));
  $a = sha1_safe_add($a, 0x5A827999);
  $a = sha1_safe_add($a, sha1_rol($data[$v] ,5));
  $data[$z] = sha1_safe_add($data[$z], $a);
  $data[$w] = sha1_rol($data[$w], 30);
}

function sha1_R2(&$block, &$data, $v, $w, $x, $y, $z, $i)
{
  $a = sha1_safe_add($data[$w] ^ $data[$x] ^ $data[$y], sha1_blk($block, $i));
  $a = sha1_safe_add($a, 0x6ED9EBA1);
  $a = sha1_safe_add($a, sha1_rol($data[$v] ,5));
  $data[$z] = sha1_safe_add($data[$z], $a);
  $data[$w] = sha1_rol($data[$w], 30);
}

function sha1_R3(&$block, &$data, $v, $w, $x, $y, $z, $i)
{
  $a = sha1_safe_add((($data[$w] | $data[$x]) & $data[$y]) | ($data[$w] & $data[$x]), sha1_blk($block, $i));
  $a = sha1_safe_add($a, 0x8F1BBCDC);
  $a = sha1_safe_add($a, sha1_rol($data[$v] ,5));
  $data[$z] = sha1_safe_add($data[$z], $a);
  $data[$w] = sha1_rol($data[$w], 30);
}

function sha1_R4(&$block, &$data, $v, $w, $x, $y, $z, $i)
{
  $a = sha1_safe_add($data[$w] ^ $data[$x] ^ $data[$y], sha1_blk($block, $i));
  $a = sha1_safe_add($a, 0xCA62C1D6);
  $a = sha1_safe_add($a, sha1_rol($data[$v] ,5));
  $data[$z] = sha1_safe_add($data[$z], $a);
  $data[$w] = sha1_rol($data[$w], 30);
}

function sha1_create()
{
  $ret['state'] = array(0x67452301, 0xEFCDAB89, 0x98BADCFE, 0x10325476, 0xC3D2E1F0);
  $ret['count'] = 0;
  $ret['blockIndex'] = 0;
  for($i = 0; $i < 16; $i++) $ret['block'][$i] = 0;
  return $ret;
}

function sha1_update(&$sha1, $content)
{
  for($i = 0; $i < strlen($content); $i++) sha1_add_byte($sha1, ord($content{$i}));
}

function sha1_add_byte(&$sha1, $b)
{
  $mask = ($sha1['blockIndex'] & 3) << 3;
  $sha1['count'] += 8;
  $sha1['block'][$sha1['blockIndex'] >> 2] &= ~(0xff << $mask);
  $sha1['block'][$sha1['blockIndex'] >> 2] |= ($b & 0xff) << $mask;
  $sha1['blockIndex']++;
  if($sha1['blockIndex'] == 64)
  {
    $dd = $sha1['state'];
    sha1_R0($sha1['block'], $dd,0,1,2,3,4, 0);sha1_R0($sha1['block'], $dd,4,0,1,2,3, 1);sha1_R0($sha1['block'], $dd,3,4,0,1,2, 2);sha1_R0($sha1['block'], $dd,2,3,4,0,1, 3);
    sha1_R0($sha1['block'], $dd,1,2,3,4,0, 4);sha1_R0($sha1['block'], $dd,0,1,2,3,4, 5);sha1_R0($sha1['block'], $dd,4,0,1,2,3, 6);sha1_R0($sha1['block'], $dd,3,4,0,1,2, 7);
    sha1_R0($sha1['block'], $dd,2,3,4,0,1, 8);sha1_R0($sha1['block'], $dd,1,2,3,4,0, 9);sha1_R0($sha1['block'], $dd,0,1,2,3,4,10);sha1_R0($sha1['block'], $dd,4,0,1,2,3,11);
    sha1_R0($sha1['block'], $dd,3,4,0,1,2,12);sha1_R0($sha1['block'], $dd,2,3,4,0,1,13);sha1_R0($sha1['block'], $dd,1,2,3,4,0,14);sha1_R0($sha1['block'], $dd,0,1,2,3,4,15);
    sha1_R1($sha1['block'], $dd,4,0,1,2,3,16);sha1_R1($sha1['block'], $dd,3,4,0,1,2,17);sha1_R1($sha1['block'], $dd,2,3,4,0,1,18);sha1_R1($sha1['block'], $dd,1,2,3,4,0,19);
    sha1_R2($sha1['block'], $dd,0,1,2,3,4,20);sha1_R2($sha1['block'], $dd,4,0,1,2,3,21);sha1_R2($sha1['block'], $dd,3,4,0,1,2,22);sha1_R2($sha1['block'], $dd,2,3,4,0,1,23);
    sha1_R2($sha1['block'], $dd,1,2,3,4,0,24);sha1_R2($sha1['block'], $dd,0,1,2,3,4,25);sha1_R2($sha1['block'], $dd,4,0,1,2,3,26);sha1_R2($sha1['block'], $dd,3,4,0,1,2,27);
    sha1_R2($sha1['block'], $dd,2,3,4,0,1,28);sha1_R2($sha1['block'], $dd,1,2,3,4,0,29);sha1_R2($sha1['block'], $dd,0,1,2,3,4,30);sha1_R2($sha1['block'], $dd,4,0,1,2,3,31);
    sha1_R2($sha1['block'], $dd,3,4,0,1,2,32);sha1_R2($sha1['block'], $dd,2,3,4,0,1,33);sha1_R2($sha1['block'], $dd,1,2,3,4,0,34);sha1_R2($sha1['block'], $dd,0,1,2,3,4,35);
    sha1_R2($sha1['block'], $dd,4,0,1,2,3,36);sha1_R2($sha1['block'], $dd,3,4,0,1,2,37);sha1_R2($sha1['block'], $dd,2,3,4,0,1,38);sha1_R2($sha1['block'], $dd,1,2,3,4,0,39);
    sha1_R3($sha1['block'], $dd,0,1,2,3,4,40);sha1_R3($sha1['block'], $dd,4,0,1,2,3,41);sha1_R3($sha1['block'], $dd,3,4,0,1,2,42);sha1_R3($sha1['block'], $dd,2,3,4,0,1,43);
    sha1_R3($sha1['block'], $dd,1,2,3,4,0,44);sha1_R3($sha1['block'], $dd,0,1,2,3,4,45);sha1_R3($sha1['block'], $dd,4,0,1,2,3,46);sha1_R3($sha1['block'], $dd,3,4,0,1,2,47);
    sha1_R3($sha1['block'], $dd,2,3,4,0,1,48);sha1_R3($sha1['block'], $dd,1,2,3,4,0,49);sha1_R3($sha1['block'], $dd,0,1,2,3,4,50);sha1_R3($sha1['block'], $dd,4,0,1,2,3,51);
    sha1_R3($sha1['block'], $dd,3,4,0,1,2,52);sha1_R3($sha1['block'], $dd,2,3,4,0,1,53);sha1_R3($sha1['block'], $dd,1,2,3,4,0,54);sha1_R3($sha1['block'], $dd,0,1,2,3,4,55);
    sha1_R3($sha1['block'], $dd,4,0,1,2,3,56);sha1_R3($sha1['block'], $dd,3,4,0,1,2,57);sha1_R3($sha1['block'], $dd,2,3,4,0,1,58);sha1_R3($sha1['block'], $dd,1,2,3,4,0,59);
    sha1_R4($sha1['block'], $dd,0,1,2,3,4,60);sha1_R4($sha1['block'], $dd,4,0,1,2,3,61);sha1_R4($sha1['block'], $dd,3,4,0,1,2,62);sha1_R4($sha1['block'], $dd,2,3,4,0,1,63);
    sha1_R4($sha1['block'], $dd,1,2,3,4,0,64);sha1_R4($sha1['block'], $dd,0,1,2,3,4,65);sha1_R4($sha1['block'], $dd,4,0,1,2,3,66);sha1_R4($sha1['block'], $dd,3,4,0,1,2,67);
    sha1_R4($sha1['block'], $dd,2,3,4,0,1,68);sha1_R4($sha1['block'], $dd,1,2,3,4,0,69);sha1_R4($sha1['block'], $dd,0,1,2,3,4,70);sha1_R4($sha1['block'], $dd,4,0,1,2,3,71);
    sha1_R4($sha1['block'], $dd,3,4,0,1,2,72);sha1_R4($sha1['block'], $dd,2,3,4,0,1,73);sha1_R4($sha1['block'], $dd,1,2,3,4,0,74);sha1_R4($sha1['block'], $dd,0,1,2,3,4,75);
    sha1_R4($sha1['block'], $dd,4,0,1,2,3,76);sha1_R4($sha1['block'], $dd,3,4,0,1,2,77);sha1_R4($sha1['block'], $dd,2,3,4,0,1,78);sha1_R4($sha1['block'], $dd,1,2,3,4,0,79);
    $sha1['state'][0] = sha1_safe_add($sha1['state'][0], $dd[0]);
    $sha1['state'][1] = sha1_safe_add($sha1['state'][1], $dd[1]);
    $sha1['state'][2] = sha1_safe_add($sha1['state'][2], $dd[2]);
    $sha1['state'][3] = sha1_safe_add($sha1['state'][3], $dd[3]);
    $sha1['state'][4] = sha1_safe_add($sha1['state'][4], $dd[4]);
    $sha1['blockIndex'] = 0;
  }
}

function sha1_finish(&$sha1)
{
  for($i = 0; $i < 8; $i++)
    $bits[$i] = sha1_zero_fill($sha1['count'], (7 - $i) << 3) & 0xff;

  sha1_add_byte($sha1, 128);
  while($sha1['blockIndex'] != 56)
    sha1_add_byte($sha1, 0);
  for($i = 0; $i < 8; ++$i) sha1_add_byte($sha1, $bits[$i]);

  $raw = '';
  for($i = 0; $i < 20; $i++)
    $raw .= chr(sha1_zero_fill($sha1['state'][$i>>2], ((3-($i & 3)) << 3)) & 0xff);
  return $raw;
}

?>